/*
 *  GPSdo - Software for GPSDO
 *  (c) Ad Sekeris 2019..
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 */

#include "Arduino.h"
#include <SPI.h>
#include <math.h>

#include "configuration.h"
// Select the proper language file for displayed messages as set in configuration.h
#if (LANG_GB)
  #include "language_gb.h"
#elif (LANG_NL)
  #include "language_nl.h"
#endif

#if (LCD_2004)
  #include <LiquidCrystal_I2C.h>
  LiquidCrystal_I2C lcd(LCD_I2C_addr,20,4);  // set the LCD address for a 20 chars and 4 line display
#endif

// Load the library for EEprom support
#include <EEPROM.h>

// Load the library for the rotary encoder and pushbutton
#include <SimpleRotary.h>
// Pin A, Pin B, Button Pin
SimpleRotary rotary(2,3,4);

// Load the library for local and summer/winter time
#include <Timezone.h>   // https://github.com/JChristensen/Timezone

// Rules per timezone. Extra rules can be defined as long as memory is available.
// See the library documentation how to build new rules
#if (LOCAL_TIME_ZONE_CE)
// Central European Time (Frankfurt, Paris, Amsterdam) utc+1
TimeChangeRule CEST = {"CEST", Last, Sun, Mar, 2, 120};     // Central European Summer Time   - utc+2
TimeChangeRule CET = {"CET ", Last, Sun, Oct, 3, 60};       // Central European Standard Time - utc+1
Timezone CE(CEST, CET);
#elif (LOCAL_TIME_ZONE_EE)
// Eastern European Time (Athens) utc+2
TimeChangeRule EEST = {"EEST", Last, Sun, Mar, 2, 180};     // Central European Summer Time   - utc+3
TimeChangeRule EET = {"EET ", Last, Sun, Oct, 3, 120};       // Central European Standard Time - utc+2
Timezone EE(EEST, EET);
#elif (LOCAL_TIME_ZONE_UK)
// United Kingdom (London, Belfast) utc
TimeChangeRule BST = {"BST", Last, Sun, Mar, 1, 60};        // British Summer Time  - utc+1
TimeChangeRule GMT = {"GMT", Last, Sun, Oct, 2, 0};         // Standard Time        - utc
Timezone UK(BST, GMT);
#endif
TimeChangeRule *tcr;        // pointer to the time change rule, use to get TZ abbrev

// Load the library for the Si5351
#include <si5351.h>   // Bibliotheek 'Etherkit Si5351' 
Si5351 si5351;

// Load the functions for the ADF4351
#define SCLK 52  // SPI bus connections
#define MOSI 51
#define SS 47  //SPI-SS om de ADF435x te selecteren
//#define xtal  25000000
#define xtal  10000000
#include "adf4351_func.h" // synth library
ADF4351 dds;  // instantiate "dds" object

// Load the library for proper GPS decoding
#include <TinyGPS++.h>  // https://github.com/mikalhart/TinyGPSPlus
TinyGPSPlus gps;
TinyGPSCustom satsInView(gps, "GPGSV", 3);         // $GPGSV sentence, third element
TinyGPSCustom satFixMode(gps, "GPGSA", 2);         // $GPGSA sentence, second element
TinyGPSCustom satLattitude(gps, "GPGGA", 2);         // $GPGGA sentence, second element
TinyGPSCustom satLongitude(gps, "GPGGA", 4);         // $GPGGA sentence, fourth element
TinyGPSCustom satsInUse(gps, "GPGGA", 7);         // $GPGGA sentence, seventh element
TinyGPSCustom satTime(gps, "GPRMC", 1);         // $GPRMC sentence, first element
TinyGPSCustom satDate(gps, "GPRMC", 9);         // $GPRMC sentence, nineth element

#include "config_func.h"
#include "ublox_func.h"
#include "display_func.h"

// Compatibility with older IDE versions
#if defined (ARDUINO) && ARDUINO >= 100
  #define printByte(args)  write(args);
#else
  #define printByte(args)  print(args,BYTE);
#endif

void setup () {
  Start_Display_Welcome ();
  // Small hardware test for the leds
  LedSwitch (Led1,Led1Mode,Led1Stat);  // During the setup turn the first led on
  delay(500);
  LedSwitch (Led2,Led2Mode,Led2Stat);  // During the setup turn the second led off
  delay(500);
  LedSwitch (Led3,Led3Mode,Led3Stat);  // During the setup turn the third led off
  delay(500);
  Led3Stat = SINGLE_OFF;
  LedSwitch (Led3,Led3Mode,Led3Stat);  // During the setup turn the third led off
  delay(500);
  Led2Stat = SINGLE_OFF;
  LedSwitch (Led2,Led2Mode,Led2Stat);  // During the setup turn the second led off
  analogReference(INTERNAL1V1);         //Lower the reference of the AD converter because the phase indicator is not using the full 5 volt range
  //analogReference(INTERNAL2V56);
  //analogReference(DEFAULT);
  Get_Start_Values();
  if (USB_DEBUG or NmeaUSBOn == 1) Serial.begin (NmeaUSBBaud);
  if (NmeaRS232On == 1 && RS232chip == 1) Serial2.begin (NmeaRS232Baud);  
  Serial1.begin(9600); //Gps_Baud_Rate = Find_Gps_Baud (settingsArray); // Sort of auto baud rate 
//  if (Gps_Baud_Rate == 9600) configureUblox (settingsArray);
  configureUblox (settingsArray);
//  else GpsHardwareFail(); // GPS failed
  if (!GpsStatus) GpsHardwareFail(); // GPS failed
  Led1Stat = SINGLE_OFF;
  LedSwitch (Led1,Led1Mode,Led1Stat);  // After GPS init success turn the first led off
  //BuildScreen0 ();
  si5351.init(SI5351_CRYSTAL_LOAD_0PF, Si5351_reference, 0);
  if (Si5351chip == 1) {
    Si0Freq = (Si0MHertz * 100000000ull)+(Si0kHertz*100000)+(Si0Hertz*100)+Si0mHertz;
    Si1Freq = (Si1MHertz * 100000000ull)+(Si1kHertz*100000)+(Si1Hertz*100)+Si1mHertz;
    Si2Freq = (Si2MHertz * 100000000ull)+(Si2kHertz*100000)+(Si2Hertz*100)+Si2mHertz;
    si5351.set_freq(Si0Freq, SI5351_CLK0);
    if (Si0Level == 0) {
      si5351.output_enable(SI5351_CLK0, 0); 
    }
    else {
      si5351.drive_strength(SI5351_CLK0, Si0Level - 1);
      si5351.output_enable(SI5351_CLK0, 1);       
    }
    si5351.set_freq(Si1Freq, SI5351_CLK1);
    if (Si1Level == 0) {
      si5351.output_enable(SI5351_CLK1, 0); 
    }
    else {
      si5351.drive_strength(SI5351_CLK1, Si1Level - 1);
      si5351.output_enable(SI5351_CLK1, 1);       
    }
    si5351.set_freq(Si2Freq, SI5351_CLK2);
    if (Si2Level == 0) {
      si5351.output_enable(SI5351_CLK2, 0); 
    }
    else {
      si5351.drive_strength(SI5351_CLK2, Si2Level - 1);
      si5351.output_enable(SI5351_CLK2, 1);       
    }
  }
  else {
      si5351.output_enable(SI5351_CLK0, 0); 
      si5351.output_enable(SI5351_CLK1, 0); 
      si5351.output_enable(SI5351_CLK2, 0);     
  }
  si5351.update_status();

  dds.init();   // init dds

  if (Adf4351chip == 1) {
    if (AdfLevel == 0) dds.setRFOut(0);
    else {
      dds.setRFPower(AdfLevel-1);
      dds.setRFOut(1);
    }
    AdfFreq = (AdfMHz*1000000ull)+(AdfkHz*1000ull);
    dds.setFreq(AdfFreq, AdfStep);
    dds.update();
  }
  LcdTotTime = millis();
  AnalogTime = millis();
  #if (LCD_2004)
    LcdmeterInit();
  #endif
}

void loop () {
  bool newData = false;
  
  if(Serial1.available())
  {
    char c = Serial1.read();  
    // Write every character to GPS for decoding.
    if (gps.encode(c)) newData = true; // Did a new valid sentence come in?     
    if (!USB_DEBUG) {if (NmeaUSBOn==1)Serial.write(c);}
    if (NmeaRS232On==1) Serial2.write(c);
  }
  if(Serial.available())
  {
    // Read incoming char and discard
    char c1 = Serial1.read();  
  }
  if(Serial2.available())
  {
    // Read incoming char and discard
    char c2 = Serial1.read();  
  }
  if (newData)
  {
    if (satFixMode.isUpdated()) {
      UpdateSatFixMode();
    }
    if (satsInUse.isUpdated()) {
      if (ScreenNew<=4)
      UpdateSatsInUse();
    }
    if (satsInView.isUpdated()) {
      if (ScreenNew<=4)
      UpdateSatsInView();
    }
    if (satLattitude.isUpdated()) {
      if (ScreenNew==2)
      UpdateLattitude();
    }
    if (satLongitude.isUpdated()) {
      if (ScreenNew==2)
      UpdateLongitude();
    }
    if (satDate.isUpdated()) {
      UpdateUtcDate();
      if (ScreenNew==1)
      UpdateLocalDate();  
    }
    if (satTime.isUpdated()) {
      UpdateUtcTime();
      if (ScreenNew==1)
      UpdateLocalTime();  
    }
    newData = false;
  } 
  if ((LatUpdate) && (LngUpdate) && (ScreenNew==2)) {
    calcLocator(NEW_LOCATOR, Lattitude, Longitude); 
    DisplayLocator();
    LatUpdate = false;
    LngUpdate = false;
  }
  // Check if the encoder has moved and select next or previous screen if not in edit mode
  EncoderPush = rotary.push();
  if (EncoderPush == 1) {
    #if (LCD_2004)
      lcd.backlight();
      LcdTotTime = millis();
    #endif
    switch (ScreenNew) {
      case 4:
        LcdMeterCalibration();
        break;
      case 5:
        SelectHardware();
        break;
      case 6:
        SelectBaudrate();
        break;
      case 7:
        SelectOption();
        break;
      case 8:
        SelectSiClk0();
        break;
      case 9:
        SelectSiClk1();
        break;
      case 10:
        SelectSiClk2();
        break;
      case 11:
        SelectAdfClk();
        break;
    }
  }
  EncoderRotate = rotary.rotate();
  if (EncoderRotate != 0) {
    #if (LCD_2004)
      lcd.backlight();
      LcdTotTime = millis();
    #endif
    if (!EditMode) {
      if (EncoderRotate == 1){
        ScreenNew += 1;
        if (ScreenNew > MaxScreen) ScreenNew=0;
        ScreenState();
      }
      else{
        ScreenNew -= 1;
        if (ScreenNew < 0 || ScreenNew == 0xFF) ScreenNew = MaxScreen;
        ScreenState();    
      }
    }
    else {
      switch (ScreenNew) {
        case 4:
          RotateLCDCalibration(EncoderRotate);
          break;
        case 5:
          ToggleSelectedHardware();
          break;
        case 6:
          switch (BaudrateSelector) {
          case 1:
            ToggleUSBNmea();
            break;
          case 2:
            SelectUSBBaudRate(EncoderRotate);
            break;
          case 3:
            ToggleRS232Nmea();
            break;
          case 4:  
            SelectRS232audRate(EncoderRotate);
            break;
          }
          break;
        case 7:
          SelectLcdTot(EncoderRotate);
          break;
        case 8:
          switch (SiClk0Selector) {
          case 1:
            Si0Level = RotateClkLevel (EncoderRotate, Si0Level, 8, 1, 12, 1 ,0 , 2);
            if (Si0Level == 0) {
              si5351.output_enable(SI5351_CLK0, 0); 
            }
            else if (Si0Level >= 1){
              si5351.drive_strength(SI5351_CLK0, Si0Level - 1);
              si5351.output_enable(SI5351_CLK0, 1); 
            } 
            break;
          case 2:
            Si0MHertz = RotateMHzFrequency (EncoderRotate, Si0MHertz, 0, 2, 3, 3, 1, 60);
            break;
          case 3:
            Si0kHertz = RotatekHzFrequency (EncoderRotate, Si0kHertz, 4, 2, 7, 3, 0, 999);
            break;
          case 4:  
            Si0Hertz = RotatekHzFrequency (EncoderRotate, Si0Hertz, 8, 2, 11, 3, 0, 999);
            break;
          case 5:  
            Si0mHertz = RotatemHzFrequency (EncoderRotate, Si0mHertz, 12, 2, 14, 3, 0, 99);
            break;
          }
          break;
        case 9:
          switch (SiClk1Selector) {
          case 1:
            Si1Level = RotateClkLevel (EncoderRotate, Si1Level, 8, 1, 12, 1 ,0 , 2);
            if (Si1Level == 0) {
              si5351.output_enable(SI5351_CLK1, 0); 
            }
            else if (Si1Level >= 1){
              si5351.drive_strength(SI5351_CLK1, Si1Level - 1);
              si5351.output_enable(SI5351_CLK1, 1); 
            } 
            break;
          case 2:
            Si1MHertz = RotateMHzFrequency (EncoderRotate, Si1MHertz, 0, 2, 3, 3, 1, 60);
            break;
          case 3:
            Si1kHertz = RotatekHzFrequency (EncoderRotate, Si1kHertz, 4, 2, 7, 3, 0, 999);
            break;
          case 4:  
            Si1Hertz = RotatekHzFrequency (EncoderRotate, Si1Hertz, 8, 2, 11, 3, 0, 999);
            break;
          case 5:  
            Si1mHertz = RotatemHzFrequency (EncoderRotate, Si1mHertz, 12, 2, 14, 3, 0, 99);
            break;
          }
          break;
        case 10:
          switch (SiClk2Selector) {
          case 1:
            Si2Level = RotateClkLevel (EncoderRotate, Si2Level, 8, 1, 12, 1 ,0 , 2);
            if (Si2Level == 0) {
              si5351.output_enable(SI5351_CLK2, 0); 
            }
            else if (Si2Level >= 1){
              si5351.drive_strength(SI5351_CLK2, Si2Level - 1);
              si5351.output_enable(SI5351_CLK2, 1); 
            } 
            break;
          case 2:
            Si2MHertz = RotateMHzFrequency (EncoderRotate, Si2MHertz, 0, 2, 3, 3, 1, 60);
            break;
          case 3:
            Si2kHertz = RotatekHzFrequency (EncoderRotate, Si2kHertz, 4, 2, 7, 3, 0, 999);
            break;
          case 4:  
            Si2Hertz = RotatekHzFrequency (EncoderRotate, Si2Hertz, 8, 2, 11, 3, 0, 999);
            break;
          case 5:  
            Si2mHertz = RotatemHzFrequency (EncoderRotate, Si2mHertz, 12, 2, 14, 3, 0, 99);
            break;
          }
        break;
      case 11:
          switch (AdfClkSelector) {
          case 1:
            RotateAdfLevel (EncoderRotate);
            if (AdfLevel == 0) {
              dds.setRFOut(0);
              dds.update();
            }
            else if (AdfLevel >= 1){
              dds.setRFOut(1);
              dds.setRFPower(AdfLevel-1);
              AdfFreq = (AdfMHz*1000000ull)+(AdfkHz*1000ull);
              dds.setFreq(AdfFreq, AdfStep);
              dds.update();
            } 
            break;
          case 2:
            RotateAdfMHz (EncoderRotate);
            break;
          case 3:
            RotateAdfkHz (EncoderRotate);
            break;
          }
          break;
      }
    } 
  }
  UpdateRunningTime ();
  UpdateFixTime ();
  ScreenState ();
  if ((millis()-Led2Blink)>=25){ 
    Led2Stat = SINGLE_OFF;
    LedSwitch (Led2,Led2Mode,SINGLE_OFF);  // GPS data processed. second led off
  }
  // blink first led with 1Hz if no sattelite fix
  if (NewFixMode <= 1) {
    #if (LCD_2004)
      lcd.backlight();
      LcdTotTime = millis();
    #endif
    if ((millis()-Led1Blink) >= 500){
      Led1Blink = millis();
      if (Led1Stat == SINGLE_ON) {
        Led1Stat = SINGLE_OFF;            
        LedSwitch (Led1,Led1Mode,Led1Stat);
      }
      else {
        Led1Stat = SINGLE_ON;            
        LedSwitch (Led1,Led1Mode,Led1Stat);
      }
    }
  }
  else if (Led1Stat == SINGLE_ON) {
    Led1Stat = SINGLE_OFF;            
    LedSwitch (Led1,Led1Mode,Led1Stat);
  }
  switch (LcdTot){
    case 0:
      break;
    case 1:
      #if (LCD_2004)
        if ((millis()-LcdTotTime)>=300000ul)lcd.noBacklight();                  // Lcd backlight off
      #endif
      break;
    case 2:
      #if (LCD_2004)
        if ((millis()-LcdTotTime)>=600000ul)lcd.noBacklight();                  // Lcd backlight off
      #endif
      break;
    case 3:
      #if (LCD_2004)
        if ((millis()-LcdTotTime)>=900000ul)lcd.noBacklight();                  // Lcd backlight off
      #endif
      break;
  } 
  if ((millis()-AnalogTime)>=200){ 
    AnalogTime = millis();
    AnInput0=analogRead(0);
    if (AnInput0 > WeighedHigh) WeighedHigh = (0.999 * WeighedHigh) + (0.001 * AnInput0);
    NeedlePosition();
    #if (LCD_2004)
      if (ScreenNew == 4) {
        LCD2004Meter();
      }
    #endif
  }
// Initial warming up time before start of checks
//   if (Fix_Hours>0 || Fix_Days>0){
  if (Fix_Minutes>=15 || Fix_Hours>0 || Fix_Days>0){
    // check for stable meter 
//    if ((millis()-MeterDwellTime)>=600000 && NewFixMode > 1){   // 10 minutes
//    if ((millis()-MeterDwellTime)>=300000 && NewFixMode > 1){   // 5 minutes
//    if ((millis()-MeterDwellTime)>=180000 && NewFixMode > 1){   // 3 minutes
//    if ((millis()-MeterDwellTime)>=180000 && NewFixMode > 1){   // 2 minutes
    if ((millis()-MeterDwellTime)>=60000 && NewFixMode > 1){   // 1 minute
//      if (OldNeedlePos == NeedlePos) {
      if (OldNeedlePos >= NeedlePos-1 && OldNeedlePos <= NeedlePos+1) {
        Led3Stat = SINGLE_ON;            
        LedSwitch (Led3,Led3Mode,Led3Stat);        
      }
      else {
        Led3Stat = SINGLE_OFF;            
        LedSwitch (Led3,Led3Mode,Led3Stat);        
      }      
      OldNeedlePos = NeedlePos;
      MeterDwellTime = millis();
    }
  }
  else {
    MeterDwellTime = millis();
    Led3Stat = SINGLE_OFF;            
    LedSwitch (Led3,Led3Mode,Led3Stat);
  }
}
